local AddonName, GGL = ...

local AceGUI = LibStub("AceGUI-3.0")
local Templates = GGL.Templates
local L = GGL.Locale

local Challenge = {}
GGL.GUI.Challenge = Challenge

local GetMoneyText = GGL.GetMoneyText
local CreateLine = GGL.GUI.CreateLine

local MONEY_WITH_NAME_FORMAT = "|cffFFF569%s|r"
local MONEY_WITH_RANK_FORMAT = "|cffFFF569%d|r"
local MONEY_WITH_NAME_AND_RANK_FORMAT = "|cffFFF569%d - %s|r"
local MONEY_WITH_NAME_AND_RANK1_FORMAT = "|cffFFF5691/%d - %s|r"

local TOP_ITEMS_LIST_SIZE = GGL.CHALLENGE_TT_MAX_TOP_ITEMS

local WIDTH_V, HEIGHT_V = 300, 350
local WIDTH_M, HEIGHT_M = 300, 900

local GetPlayerTemplate = Templates.Sync.GetPlayerFromSync

local function SetMoneyText(self, money, playerName, rank, attendees)
    if not money or money <= 0 then
        self:SetText("-")
        if self.name then
            self.name:SetText("-")
        end
    else
        if self.name then
            if playerName and playerName ~= "" then
                if attendees then
                    self.name:SetFormattedText(MONEY_WITH_NAME_AND_RANK1_FORMAT, attendees, playerName)
                elseif rank then
                    self.name:SetFormattedText(MONEY_WITH_NAME_AND_RANK_FORMAT, rank, playerName)
                else
                    self.name:SetFormattedText(MONEY_WITH_NAME_FORMAT, playerName)
                end
            else
                if rank then
                    self.name:SetFormattedText(MONEY_WITH_RANK_FORMAT, rank)
                else
                    self.name:SetText("")
                end
            end
        end
        self:SetText(GetMoneyText(money))
    end
end

-- ####################
-- Challenge frame

-- EVENTS
--      "ChallengeList_OnListSort"        (challengeFrame)        -- List finished updating
-- ####################

local function CharContainer_OnListSort(self)
    Challenge.frame:RefreshGoldCount()
    Challenge.frame:FireCallbackEvent("ChallengeList_OnListSort", Challenge.frame)
end

local function ExportButton_OnClick(self)
    GGL.Export:FillExportWindow(Challenge.frame:ExportPlayerList(), "csv")
end

local function AttendeesButton_OnClick(self)
    if not Challenge.frame.session then return end
    Challenge:ShowInviteFrame(Challenge.frame.session)
end

local FRAME_FUNCTIONS = {
    ["Clear"] = function(self)
        -- clear lists
        self:SetUpdateInterval()
        self:ClearPlayerList()

        -- clear session
        self:SetSession()

        -- set gold
        self:SetGoldCount(0)

        self:RefreshAttendees()

        self:ShowExportButton(false)

        -- reset itemList
        wipe(self.mergeIDSave)
    end,

    --- Sets the linked session to the frame
    -- @param       sessionTemplate         <sessionTemplate>
    ["SetSession"] = function(self, sessionTemplate)
        self.session = sessionTemplate
    end,

    ["UpdateMergeList"] = function(self)
        if not self.session then return end

        for playerSyncID, row in pairs(self.RowList) do
            local mergeID = GetPlayerTemplate(playerSyncID)
            if mergeID then mergeID = self.session:GetPlayerMergeID(mergeID) end
            if row and mergeID and mergeID ~= 0 and mergeID ~= playerSyncID then

                -- mergeID found
                if self.RowList[mergeID] then
                    -- the mergeID entry exists so just merge them
                    local validMerge = row:MergeInto(self.RowList[mergeID])
                    self.charContainer:ReleaseSingleChildren(row)
                    self.RowList[playerSyncID] = nil
                    -- update Player
                    self.RowList[mergeID]:SetPlayer(Templates.Sync.GetPlayerFromSync(playerSyncID, true))
                    self.RowList[mergeID]:SetInviteState(Templates.Sync.GetPlayerFromSync(playerSyncID, true))
                elseif not self.RowList[mergeID] then
                    -- mergeID entry not in list, we just "rename" it
                    self.RowList[mergeID] = self.RowList[playerSyncID]
                    -- remove old one
                    self.RowList[playerSyncID] = nil
                end
            end
        end

        -- force resort
        self.charContainer:SortChildrenList()
    end,

    --- Removes all players from the list
    ["ClearPlayerList"] = function(self)
        self.charContainer:ReleaseChildren()
        wipe(self.RowList)
    end,

    --- Sets the update interval for the player list
    -- @param       interval        <number> time in seconds
    ["SetUpdateInterval"] = function(self, interval)
        self.charContainer:SetRefreshInterval(interval)
    end,

    --- Force a update for the player list
    ["UpdateList"] = function(self)
        self.charContainer:SortChildrenList()
    end,

    ["SetGoldCount"] = function(self, copper, add)
        self.moneySumCount = add and ((self.moneySumCount or 0) + copper) or (copper or 0)
        self.moneySum:SetMoneyText(self.moneySumCount)
    end,

    --- Refresh the goldcount of all players
    ["RefreshGoldCount"] = function(self)
        local children = self.charContainer.children
        local copper = 0
        for i = 1, #children do
            copper = copper + (children[i]:GetMoney() or 0)
        end
        self:SetGoldCount(copper, false)
    end,

    --- Returns number of players in the challenge list
    ["GetAttendees"] = function(self)
        return #self.charContainer.children or 0
    end,

    --- Refresh the attendees text on top of the frame
    ["RefreshAttendees"] = function(self)
        --self.attendees:SetFormattedText(L["Attendees: |cffffd70a%d|r"], self:GetAttendees())
        self.attendeesNum:SetText(self:GetAttendees())
    end,

    --- Returns the row of a player from the challenge list
    -- @returns playerChallengeLine
    ["GetPlayerRow"] = function(self, playerTemplate)
        if not playerTemplate then return end
        return self.RowList[self:GetPlayerListKey(playerTemplate)]
    end,

    --- Returns playerItemListKey
    -- @param       playerTemplate      <playerTemplate>
    ["GetPlayerListKey"] = function(self, playerTemplate)
        if not playerTemplate then return end
        if self.RowList[playerTemplate] then
            return playerTemplate
        else
            return self.session:GetPlayerMergeID(playerTemplate) or playerTemplate.syncID
        end
    end,

    --- Adds a player into the challenge list
    -- @param       playerTemplate      <playerTemplate>
    ["AddPlayer"] = function(self, playerTemplate)
        if not playerTemplate then return end

        -- check if row is existing
        if self:GetPlayerRow(playerTemplate) then
            self:GetPlayerRow(playerTemplate):SetPlayer(playerTemplate)
            self:GetPlayerRow(playerTemplate):SetInviteState(self.session:GetPlayerInviteState(playerTemplate))
            return
        end

        -- check for mergeID
        local mergeID = self.session:GetPlayerMergeID(playerTemplate)
        if mergeID and self.RowList[mergeID] then
            -- mergeID entry exists so we only set the active player
            self.RowList[mergeID]:SetPlayer(playerTemplate)
            self.RowList[mergeID]:SetInviteState(self.session:GetPlayerInviteState(playerTemplate))
            return
        end

        -- get listKey
        local playerListKey = self:GetPlayerListKey(playerTemplate)

        -- create a new row
        local row = AceGUI:Create("GoldGoblinLootGUIWidget-ChallengeLine")
        row:SetSessionResetTable(self.session:InstanceReset_GetData())
        row:SetFullWidth(true)
        row:SetSession(self.session)
        row:SetPlayer(playerTemplate)
        row:SetInviteState(self.session:GetPlayerInviteState(playerTemplate))
        row:SetToolTipPoint("BOTTOM_X", self.frame)
        self.charContainer:AddChild(row)

        -- add into rowList
        self.RowList[playerListKey] = row

        -- refresh attendees
        self:RefreshAttendees()
    end,

    --- Sets the inviteState for a player
    -- @param       player      <playerTemplate>
    -- @param       state       <GGL.INVITE_KEY_LIST>
    ["SetPlayerInviteState"] = function(self, player, state)
        if not player or not self:GetPlayerRow(player) then return end
        self:GetPlayerRow(player):SetInviteState(state)
    end,

    --- Unregisters updates for all players (like name changes)
    ["UnregisterPlayerUpdates"] = function(self)
        local children = self.charContainer.children
        for i = 1, #children do
            children[i]:UnregisterPlayerUpdates()
        end
    end,

    --- Sets the gold for a player
    -- This not updates immediately and needs a list refresh
    -- @param       player      <playerTemplate>
    -- @param       money       <number> copper
    -- @param       replace     <bool> replace the money value of a player with the given
    ["SetPlayerGold"] = function(self, player, money, replace)
        if not self:GetPlayerRow(player) then return end

        self:GetPlayerRow(player):SetMoney(money, not replace, false)
    end,

    --- Export players as table
    ["ExportPlayerList"] = function(self)
        -- self.charContainer.children
        local exportTab = {
            header = {"Place", "PlayerName", "GoldValue", "TopItemName", "TopItemID", "TopItemValue", "RewardID"}
        }

        for place, child in ipairs(self.charContainer.children) do
            local childT = {}

            -- basic data
            childT[1] = place
            childT[2] = child:GetName()
            childT[3] = child:GetMoney()

            -- topItems
            -- itemID, quantity, moneyValue, sourceItemID, mainSource, subSource
            local topItemList = child.itemList and child.itemList[1] or false

            if topItemList then
                local itemName, itemLink = GetItemInfo(topItemList[1])
                childT[4] = itemName
                childT[5] = topItemList[1]
                childT[6] = topItemList[3]
            else
                childT[4] = NONE
                childT[5] = 0
                childT[6] = 0
            end

            -- get rewardID
            childT[7] = child._player and child._player:GetGGRewardID() or 0

            -- set entry
            exportTab[#exportTab+1] = childT
        end

        return exportTab
    end,

    --- Shows a export button over the timer
    ["ShowExportButton"] = function(self, state)
        if state then
            --self.exportButton.frame:Show()
        else
            --self.exportButton.frame:Hide()
        end
    end,
}

function Challenge:CreateFrame()
    if not self.frame then
        local window = AceGUI:Create("Window")
        window:SetTitle("GoldGoblinLoot - "..L["Leaderboard"])

        local frame = window.frame
        local content = window.content

        -- some frame settings
        frame:SetMinResize(WIDTH_V,HEIGHT_V)
        frame:SetMaxResize(WIDTH_V+WIDTH_M,HEIGHT_V+HEIGHT_M)
        frame:SetClampedToScreen(true)
        frame:SetToplevel(false)
        frame:SetFrameStrata("MEDIUM")
        --window settings
        window:SetWidth(WIDTH_V)
        window:SetHeight(HEIGHT_V)
        -- remove resizer lines
        window.line1:Hide()
        window.line2:Hide()

        -- ## timer
        local timerContainer = GGL.GUI:CreateBoxFrame(content)
        timerContainer:SetPoint("TOP", content, "TOP", 0, 0)
        timerContainer:SetSize(120, 27)

        local timer = AceGUI:Create("GoldGoblinLootGUIWidget-Timer")
        local timerFrame = timer.frame
        timerFrame:SetParent(content)
        timer:SetPoint("CENTER", timerContainer, 0, 0)
        timer:SetSize(20)
        timer:SetDuration(0)
        window.timer = timer
        timerContainer:SetSize(timerFrame:GetWidth()+10, 27)

        local exportButton = AceGUI:Create("Button")
        exportButton.frame:SetParent(content)
        exportButton.frame:SetPoint("TOPRIGHT", content, "TOPRIGHT", 0, 2)
        exportButton:SetText(L["Export"])
        exportButton:SetAutoWidth(true)
        exportButton:SetCallback("OnClick", ExportButton_OnClick)
        exportButton.frame:Show()
        window.exportButton = exportButton

        local attendeesNumBG = GGL.GUI:CreateBoxFrame(content)
        local attendeesNum = attendeesNumBG:CreateFontString(nil, "BACKGROUND", "GameFontHighlightLarge")
        attendeesNum:SetPoint("TOP",  timerContainer, 0, 0)
        attendeesNum:SetPoint("BOTTOM", timerContainer, 0, 0)
        attendeesNum:SetPoint("LEFT", content, 5, 0)
        attendeesNum:SetPoint("RIGHT", timerContainer, "LEFT", -2, 0)
        attendeesNum:SetJustifyH("CENTER")
        attendeesNum:SetJustifyV("MIDDLE")
        --attendeesNum:SetTextColor(1, 0.96, 0.41)
        attendeesNum:SetText(999)
        --attendeesNum:SetWidth(attendeesNum:GetUnboundedStringWidth() + 4)
        window.attendeesNum = attendeesNum

        attendeesNumBG:SetPoint("CENTER", attendeesNum, 0, 0)
        attendeesNumBG:SetWidth(attendeesNum:GetUnboundedStringWidth() + 4)
        attendeesNumBG:SetPoint("TOP", attendeesNum, 0, 0)
        attendeesNumBG:SetPoint("BOTTOM", attendeesNum, 0, 0)

        local attendees = content:CreateFontString(nil, "BACKGROUND", "GameFontHighlightSmall")
        attendees:SetPoint("TOP", attendeesNum, "BOTTOM", 0, 2)
        attendees:SetPoint("LEFT", attendeesNum, 0, 0)
        attendees:SetPoint("RIGHT",attendeesNum, 0, 0)
        attendees:SetHeight(12)
        attendees:SetJustifyH("CENTER")
        attendees:SetJustifyV("MIDDLE")
        attendees:SetText(L["Attendees"])
        window.attendees = attendees

        local attendeesButton = CreateFrame("BUTTON", nil, content)
        attendeesButton:SetAllPoints(attendeesNumBG)
        attendeesButton:SetScript("OnClick", AttendeesButton_OnClick)

        local moneySum = content:CreateFontString(nil, "BACKGROUND", "GameFontHighlightSmall")
        moneySum:SetAllPoints(attendees)
        moneySum:SetPoint("RIGHT", content, -2, 0)
        moneySum:SetJustifyH("RIGHT")
        moneySum:SetJustifyV("MIDDLE")
        moneySum.SetMoneyText = SetMoneyText
        window.moneySum = moneySum

        --charContainerBG
        local charContainerBG = GGL.GUI:CreateBorderedFrame(content)
        -- charContainer
        local charContainer = AceGUI:Create("GoldGoblinLootGUIContainer-SortedScrollFrame")
        charContainer.frame:SetParent(frame)
        charContainer:SetFullWidth(true)
        charContainer:SetPoint("TOP", timerFrame, "BOTTOM", 0, -20)
        charContainer:SetPoint("LEFT", content, 5, 0)
        charContainer:SetPoint("RIGHT", content, -2, 0)
        charContainer:SetPoint("BOTTOM", content, 0, 5)
        charContainer:SetCallback("OnListSort", CharContainer_OnListSort)
        window.charContainer = charContainer
        --window:AddChild(charContainer)
        -- charContainer / bg
        charContainerBG:SetPoint("TOPLEFT", charContainer.scrollframe, -5, 5)
        charContainerBG:SetPoint("BOTTOMRIGHT", charContainer.scrollframe, 5, -5)

        -- register functions
        for k,v in pairs(FRAME_FUNCTIONS) do
            window[k] = v
        end
        FRAME_FUNCTIONS = nil

        GGL:AddCallbackHandler(window)

        window.RowList = {}
        window.mergeIDSave = {}

        GGL.GUI:AddFrameControl("ChallengeFrame", frame, true, true)

        self.frame = window
    end
    return self.frame
end

function Challenge:Show()
    Challenge:CreateFrame():Show()
end

function Challenge:AddIntoMiniMapButton()
    GGL:RegisterMMFunction(self, self.Show)
end

function Challenge:RemoveFromMiniMapButton()
    GGL:UnregisterMMFunction(self.Show)
end

-- ####################
-- Small frame
-- ####################
local S_WIDTH_V, S_HEIGHT_V = 250, 132
local S_WIDTH_M, S_HEIGHT_M = 150, 0

local SMALL_FRAME_FUNCTIONS = {
    ["Clear"] = function(self)
        self:SetFirstPlayerData()
        self:SetUpPlayerData()
        self:SetPlayerData()
        self:SetDownPlayerData()
    end,

    ["SetFirstPlayerData"] = function(self, copper, playerName, attendees)
        self.moneyFirst.money:SetMoneyText(copper, playerName, nil, attendees)
    end,

    ["SetUpPlayerData"] = function(self, copper, playerName, rank)
        self.moneyUp.money:SetMoneyText(copper, playerName, rank)
    end,

    ["SetPlayerData"] = function(self, copper, playerName, rank)
        self.moneyCur.money:SetMoneyText(copper, playerName, rank)
    end,

    ["SetDownPlayerData"] = function(self, copper, playerName, rank)
        self.moneyDown.money:SetMoneyText(copper, playerName, rank)
    end,
}

local ICON_TEX_COORDS = {
    ["UP"] = {0, 0.25, 0, 0.5},
    ["DOWN"] = {0, 0.25, 0.5, 1},
    ["CROWN"] = 132062,
}

local function GetMoneyLineWithTexture(frame, addIcon, iconTex)
    local line = CreateFrame("Frame", nil, frame)
    line:SetHeight(15)

    local bg = frame:CreateTexture(nil, "BACKGROUND", nil, 1)
    bg:SetAllPoints(line)
    line.bg = bg

    if addIcon then
        local image = line:CreateTexture(nil, "BACKGROUND", nil, 2)
        image:SetPoint("LEFT", line, "LEFT", 1, 0)
        image:SetSize(13,13)
        if iconTex and ICON_TEX_COORDS[iconTex] then
            if type(ICON_TEX_COORDS[iconTex]) == "table" then
                image:SetTexture(GGL.MEDIA_PATH.."RankArrows")
                image:SetTexCoord(unpack(ICON_TEX_COORDS[iconTex]))
            else
                image:SetTexture(ICON_TEX_COORDS[iconTex])
            end
        else
            image:Hide()
        end
        line.image = image
    end

    local name = line:CreateFontString(nil, "BACKGROUND", "GameFontHighlightSmall")
    name:SetPoint("RIGHT", line, "RIGHT")
    if line.image then
        name:SetPoint("LEFT", line.image, "RIGHT", 1, 0)
    else
        name:SetPoint("LEFT", line, "LEFT")
    end
    name:SetJustifyH("LEFT")
    name:SetJustifyV("MIDDLE")
    name.SetMoneyText = SetMoneyText
    name:SetText("-")
    line.name = name

    local money = line:CreateFontString(nil, "BACKGROUND", "GameFontHighlightSmall")
    money:SetAllPoints(name)
    money:SetJustifyH("RIGHT")
    money:SetJustifyV("MIDDLE")
    money.SetMoneyText = SetMoneyText
    money:SetText("-")
    money.name = name
    line.money = money

    return line
end

function Challenge:CreateSmallFrame()
    if not self.smallFrame then
        local window = AceGUI:Create("Window")
        window:SetTitle("GoldGoblinLoot - Challenge")

        local frame = window.frame
        local content = window.content

        -- some frame settings
        frame:SetMinResize(S_WIDTH_V,S_HEIGHT_V)
        frame:SetMaxResize(S_WIDTH_V+S_WIDTH_M,S_HEIGHT_V+S_HEIGHT_M)
        frame:SetClampedToScreen(true)
        frame:SetToplevel(false)
        frame:SetFrameStrata("MEDIUM")
        --window settings
        window:SetWidth(S_WIDTH_V)
        window:SetHeight(S_HEIGHT_V)
        -- remove resizer lines
        window.line1:Hide()
        window.line2:Hide()

        -- timer
        local timer = AceGUI:Create("GoldGoblinLootGUIWidget-Timer")
        timer.frame:SetParent(content)
        timer:SetPoint("TOP", content, "TOP", 0, 0)
        timer:SetDuration(0)
        window.timer = timer

        local moneyFirst = GetMoneyLineWithTexture(content, true, "CROWN")
        moneyFirst:SetPoint("TOPLEFT", content, "TOPLEFT", 0, -(timer.frame:GetHeight()+5))
        moneyFirst:SetPoint("RIGHT", content, "RIGHT")
        moneyFirst.bg:SetColorTexture(1,1,1,0.2)
        window.moneyFirst = moneyFirst

        local moneyUp = GetMoneyLineWithTexture(content, true, "UP")
        moneyUp:SetPoint("TOPLEFT", moneyFirst, "BOTTOMLEFT", 0, -2)
        moneyUp:SetPoint("RIGHT", content, "RIGHT")
        moneyUp.bg:SetColorTexture(0,1,0,0.1)
        window.moneyUp = moneyUp

        local moneyCur = GetMoneyLineWithTexture(content, true)
        moneyCur:SetPoint("TOPLEFT", moneyUp, "BOTTOMLEFT", 0, -1)
        moneyCur:SetPoint("RIGHT", content, "RIGHT")
        --moneyCur.bg:SetColorTexture(0,1,0,0.1)
        window.moneyCur = moneyCur

        local moneyDown = GetMoneyLineWithTexture(content, true, "DOWN")
        moneyDown:SetPoint("TOPLEFT", moneyCur, "BOTTOMLEFT", 0, -1)
        moneyDown:SetPoint("RIGHT", content, "RIGHT")
        moneyDown.bg:SetColorTexture(1,0,0,0.1)
        window.moneyDown = moneyDown

        window.lines = {
            -- under timer
            CreateLine(content, false, {"BOTTOMLEFT", moneyFirst, "TOPLEFT", 0, 3}, {"RIGHT", content, "RIGHT"}, nil, 1),
            -- moneyLines
            CreateLine(content, false, {"TOPLEFT", moneyFirst, "BOTTOMLEFT"}, {"TOPRIGHT", moneyFirst, "BOTTOMRIGHT"}, nil, 2),
            CreateLine(content, false, {"TOPLEFT", moneyUp, "BOTTOMLEFT"}, {"TOPRIGHT", moneyUp, "BOTTOMRIGHT"}, nil, 1),
            CreateLine(content, false, {"TOPLEFT", moneyCur, "BOTTOMLEFT"}, {"TOPRIGHT", moneyCur, "BOTTOMRIGHT"}, nil, 1),
        }


        -- register functions
        for k,v in pairs(SMALL_FRAME_FUNCTIONS) do
            window[k] = v
        end
        SMALL_FRAME_FUNCTIONS = nil

        window.RowList = {}

        GGL.GUI:AddFrameControl("SmallChallengeFrame", frame, true, true)

        self.smallFrame = window
    end
    return self.smallFrame
end

function Challenge:ShowSmallFrame()
    Challenge:CreateSmallFrame():Show()
    return self.smallFrame
end

function Challenge:AddSmallIntoMiniMapButton()
    GGL:RegisterMMFunction(self, self.ShowSmallFrame)
end

function Challenge:RemoveSmallFromMiniMapButton()
    GGL:UnregisterMMFunction(self.ShowSmallFrame)
end

-- ####################
-- Invite frame
--
-- EVENTS
--      "CharContainerAccepted_OnListSort"        (challengeInviteFrame)        -- List finished updating
--      "CharContainerAccepted_OnListSort"        (challengeInviteFrame)        -- List finished updating
--      "CharContainerInvited_OnListSort"        (challengeInviteFrame)         -- List finished updating
-- ####################
local INV_WIDTH_V, INV_HEIGHT_V = 600, 500
local INV_WIDTH_M, INV_HEIGHT_M = 0, 500

local INV_CHAR_CONTAINER_WIDTH = (INV_WIDTH_V - 24 - 20 - 10) / 3 -- (fullWidth - content - between - )

local function CB_InviteFrame_PlayerUpdate(self, player)
    if not Challenge.inviteFrame then return end
    Challenge.inviteFrame:UpdatePlayer(player)
end

local function CB_InviteFrame_MergeIDUpdate(self)
    if not Challenge.inviteFrame then return end
    Challenge.inviteFrame:UpdateMergeList()
end

local function CB_InviteFrame_SessionStopped(self)
    if not Challenge.inviteFrame then return end
    Challenge.inviteFrame:Clear()
    Challenge.inviteFrame:Hide()
end

local function GetNewPlayerLine(self)
    local playerLine = AceGUI:Create("GoldGoblinLootGUIWidget-InviteLine")
    playerLine:SetFullWidth(true)
    playerLine:SetToolTipPoint("BOTTOM_X", self.frame)
    return playerLine
end

local INV_FRAME_FUNCTIONS = {
    ["Clear"] = function(self)
        self:SetSession()
    end,

    ["ClearPlayerLists"] = function(self)
        self.charContainerAccepted:ReleaseChildren()
        self.charContainerDeclined:ReleaseChildren()
        self.charContainerInvited:ReleaseChildren()

        wipe(self.playerList)
    end,

    --- Sets the linked session to the frame
    -- @param       sessionTemplate         <sessionTemplate>
    ["SetSession"] = function(self, sessionTemplate)
        if self.session == sessionTemplate then return end
        if self.session then
            self.session.inviteTemplate:UnregisterCallbackEvent("SessionTemplate_Invite_PlayerInviteStateUpdate", CB_InviteFrame_PlayerUpdate)
            self.session.inviteTemplate:UnregisterCallbackEvent("SessionTemplate_Invite_NewPlayerAdded", CB_InviteFrame_PlayerUpdate)
            self.session.inviteTemplate:UnregisterCallbackEvent("SessionTemplate_Invite_PlayerRemoved", CB_InviteFrame_PlayerUpdate)
            self.session:UnregisterCallbackEvent("SessionTemplate_Sync_PlayerMergeIDUpdate", CB_InviteFrame_MergeIDUpdate)
            self.session:UnregisterCallbackEvent("SessionTemplate_SessionStopped", CB_InviteFrame_SessionStopped)
            self.session:UnregisterCallbackEvent("SessionTemplate_SessionClosed", CB_InviteFrame_SessionStopped)
            self.session = nil
        end
        if sessionTemplate and not sessionTemplate.inviteTemplate then
            sessionTemplate = nil
        end
        if sessionTemplate then
            sessionTemplate.inviteTemplate:RegisterCallbackEvent("SessionTemplate_Invite_PlayerInviteStateUpdate", CB_InviteFrame_PlayerUpdate)
            sessionTemplate.inviteTemplate:RegisterCallbackEvent("SessionTemplate_Invite_NewPlayerAdded", CB_InviteFrame_PlayerUpdate)
            sessionTemplate.inviteTemplate:RegisterCallbackEvent("SessionTemplate_Invite_PlayerRemoved", CB_InviteFrame_PlayerUpdate)
            sessionTemplate:RegisterCallbackEvent("SessionTemplate_Sync_PlayerMergeIDUpdate", CB_InviteFrame_MergeIDUpdate)
            sessionTemplate:RegisterCallbackEvent("SessionTemplate_SessionStopped", CB_InviteFrame_SessionStopped)
            sessionTemplate:RegisterCallbackEvent("SessionTemplate_SessionClosed", CB_InviteFrame_SessionStopped)
        end

        self.session = sessionTemplate
        self:UpdateFullPlayerList()
    end,

    --- Update a player and place it into the correct list
    ["UpdatePlayer"] = function(self, playerTemplate)
        if not playerTemplate or not self.session then return end
        local playerRowData = self:GetPlayerRowData(playerTemplate)
        if not playerRowData then
            self:AddPlayer(playerTemplate)
            return
        end

        local playerLine, list, lastState = playerRowData[1], playerRowData[2], playerRowData[3]
        local curState = self.session:GetPlayerInviteState(playerTemplate)

        playerLine:SetPlayer(playerTemplate)
        playerLine:SetInviteState(curState)

        if curState ~= lastState or not list then
            if list then
                list:RemoveSingleChildren(playerLine)
            end
            if curState == GGL.INVITE_KEY_LIST.ACCEPTED then
                playerRowData[2] = self.charContainerAccepted
            elseif curState == GGL.INVITE_KEY_LIST.DECLINED then
                playerRowData[2] = self.charContainerDeclined
            else
                -- all other into invited / GGL.INVITE_KEY_LIST.INVITED
                playerRowData[2] = self.charContainerInvited
            end
            playerRowData[2]:AddChild(playerLine)
            playerRowData[2]:SortChildrenList()
            playerRowData[3] = curState
        end

        return true
    end,

    --- Adds a player into the list
    -- @param       playerTemplate      <playerTemplate>
    ["AddPlayer"] = function(self, playerTemplate)
        if not playerTemplate or not self.session then return end

        -- check if row is existing
        if self:GetPlayerRowData(playerTemplate) then
            self:UpdatePlayer(playerTemplate)
            return
        end

        -- get listKey
        local playerListKey = self:GetPlayerListKey(playerTemplate)

        -- create a new row
        local playerLine = GetNewPlayerLine(self)

        -- add into rowList
        if self.playerList[playerListKey]  then
            self.playerList[playerListKey][1] = playerLine
        else
            self.playerList[playerListKey] = { playerLine }
        end

        -- update player
        self:UpdatePlayer(playerTemplate)
    end,

    --- Returns the row data of a player
    -- @returns playerRowData { playerLine, list, state }
    ["GetPlayerRowData"] = function(self, playerTemplate)
        if not playerTemplate or not self.session then return end
        return self.playerList[self:GetPlayerListKey(playerTemplate)]
    end,

    --- Returns playerItemListKey
    -- @param       playerTemplate      <playerTemplate>
    ["GetPlayerListKey"] = function(self, playerTemplate)
        if not playerTemplate or not self.session then return end
        if self.playerList[playerTemplate] then
            return playerTemplate
        else
            return self.session:GetPlayerMergeID(playerTemplate) or playerTemplate.syncID
        end
    end,

    --- Refcreate the list
    ["UpdateFullPlayerList"] = function(self)
        self:ClearPlayerLists()

        if not self.session then return end

        for player, inviteState in self.session:IterInvitedPlayers() do
            self:AddPlayer(player)
        end
    end,

    ["UpdateMergeList"] = function(self)
        if not self.session then return end

        for playerSyncID, mergeData in pairs(self.playerList) do
            local playerLine, frame, state = mergeData[1], mergeData[2], mergeData[3]
            local mergeID = GetPlayerTemplate(playerSyncID)
            if mergeID then mergeID = self.session:GetPlayerMergeID(mergeID) end
            if playerLine and mergeID and mergeID ~= 0 and mergeID ~= playerSyncID then

                -- mergeID found
                if self.playerList[mergeID] then
                    frame:ReleaseSingleChildren(playerLine)
                    self.playerList[playerSyncID] = nil
                    -- update Player
                    self:UpdatePlayer(Templates.Sync.GetPlayerFromSync(playerSyncID, true))
                    -- restor
                    frame:SortChildrenList()
                elseif not self.playerList[mergeID] then
                    -- mergeID entry not in list, we just "rename" it
                    self.playerList[mergeID] = self.playerList[playerSyncID]
                    -- remove old one
                    self.playerList[playerSyncID] = nil
                end
            end
        end
    end,
}

local function InviteFrame_OnClose(window)
    window:Clear()
end

local function CharContainerAccepted_OnListSort(self)
    Challenge.inviteFrame:FireCallbackEvent("CharContainerAccepted_OnListSort", Challenge.inviteFrame)
end

local function CharContainerDeclined_OnListSort(self)
    Challenge.inviteFrame:FireCallbackEvent("CharContainerDeclined_OnListSort", Challenge.inviteFrame)
end

local function CharContainerInvited_OnListSort(self)
    Challenge.inviteFrame:FireCallbackEvent("CharContainerInvited_OnListSort", Challenge.inviteFrame)
end

function Challenge:CreateInviteFrame()
    if not self.inviteFrame then
        local window = AceGUI:Create("Window")
        window:SetTitle("GoldGoblinLoot - "..L["Invite list"])

        window:SetCallback("OnClose", InviteFrame_OnClose)

        local frame = window.frame
        local content = window.content

        -- some frame settings
        frame:SetMinResize(INV_WIDTH_V,INV_HEIGHT_V)
        frame:SetMaxResize(INV_WIDTH_V+INV_WIDTH_M,INV_HEIGHT_V+INV_HEIGHT_M)
        frame:SetClampedToScreen(true)
        frame:SetToplevel(false)
        frame:SetFrameStrata("MEDIUM")
        --window settings
        window:SetWidth(INV_WIDTH_V)
        window:SetHeight(INV_HEIGHT_V)

        local acceptedHeader = content:CreateFontString(nil, "BACKGROUND", "GameFontHighlightSmall")
        acceptedHeader:SetPoint("TOPLEFT", content, 0, 0)
        acceptedHeader:SetWidth(INV_CHAR_CONTAINER_WIDTH+10)
        acceptedHeader:SetHeight(12)
        acceptedHeader:SetJustifyH("CENTER")
        acceptedHeader:SetJustifyV("MIDDLE")
        acceptedHeader:SetText(L["Accepted"])
        window.acceptedHeader = acceptedHeader

        --charContainerBG
        local charContainerAcceptedBG = GGL.GUI:CreateBorderedFrame(content)
        -- charContainer
        local charContainerAccepted = AceGUI:Create("GoldGoblinLootGUIContainer-SortedScrollFrame")
        charContainerAccepted.frame:SetParent(frame)
        charContainerAccepted:SetWidth(INV_CHAR_CONTAINER_WIDTH)
        charContainerAccepted:SetPoint("TOP", acceptedHeader, "BOTTOM", 0, -7)
        charContainerAccepted:SetPoint("LEFT", content, 5, 0)
        charContainerAccepted:SetPoint("BOTTOM", content, 0, 10)
        charContainerAccepted:SetCallback("ChallengeList_OnListSort", CharContainerAccepted_OnListSort)
        charContainerAccepted:SetDescending(true)
        window.charContainerAccepted = charContainerAccepted
        -- charContainer / bg
        charContainerAcceptedBG:SetPoint("TOPLEFT", charContainerAccepted.scrollframe, -5, 5)
        charContainerAcceptedBG:SetPoint("BOTTOMRIGHT", charContainerAccepted.scrollframe, 5, -5)

        local declinedHeader = content:CreateFontString(nil, "BACKGROUND", "GameFontHighlightSmall")
        declinedHeader:SetPoint("TOPLEFT", acceptedHeader, "TOPRIGHT", 0, 0)
        declinedHeader:SetWidth(INV_CHAR_CONTAINER_WIDTH+10)
        declinedHeader:SetHeight(12)
        declinedHeader:SetJustifyH("CENTER")
        declinedHeader:SetJustifyV("MIDDLE")
        declinedHeader:SetText(L["Declined"])
        window.declinedHeader = declinedHeader

        --charContainerBG
        local charContainerDeclinedBG = GGL.GUI:CreateBorderedFrame(content)
        -- charContainer
        local charContainerDeclined = AceGUI:Create("GoldGoblinLootGUIContainer-SortedScrollFrame")
        charContainerDeclined.frame:SetParent(frame)
        charContainerDeclined:SetWidth(INV_CHAR_CONTAINER_WIDTH)
        charContainerDeclined:SetPoint("TOP", charContainerAccepted.frame, 0, 0)
        charContainerDeclined:SetPoint("LEFT", charContainerAccepted.frame, "RIGHT", 10, 0)
        charContainerDeclined:SetPoint("BOTTOM", charContainerAccepted.frame, 0, 0)
        charContainerDeclined:SetCallback("ChallengeList_OnListSort", CharContainerDeclined_OnListSort)
        charContainerDeclined:SetDescending(true)
        window.charContainerDeclined = charContainerDeclined
        -- charContainer / bg
        charContainerDeclinedBG:SetPoint("TOPLEFT", charContainerDeclined.scrollframe, -5, 5)
        charContainerDeclinedBG:SetPoint("BOTTOMRIGHT", charContainerDeclined.scrollframe, 5, -5)

        local invitedHeader = content:CreateFontString(nil, "BACKGROUND", "GameFontHighlightSmall")
        invitedHeader:SetPoint("TOPLEFT", declinedHeader, "TOPRIGHT", 0, 0)
        invitedHeader:SetWidth(INV_CHAR_CONTAINER_WIDTH+10)
        invitedHeader:SetHeight(12)
        invitedHeader:SetJustifyH("CENTER")
        invitedHeader:SetJustifyV("MIDDLE")
        invitedHeader:SetText(L["Invited"])
        window.invitedHeader = invitedHeader

        --charContainerBG
        local charContainerInvitedBG = GGL.GUI:CreateBorderedFrame(content)
        -- charContainer
        local charContainerInvited = AceGUI:Create("GoldGoblinLootGUIContainer-SortedScrollFrame")
        charContainerInvited.frame:SetParent(frame)
        charContainerInvited:SetWidth(INV_CHAR_CONTAINER_WIDTH)
        charContainerInvited:SetPoint("TOP", charContainerDeclined.frame, 0, 0)
        charContainerInvited:SetPoint("LEFT", charContainerDeclined.frame, "RIGHT", 10, 0)
        charContainerInvited:SetPoint("BOTTOM", charContainerDeclined.frame, 0, 0)
        charContainerInvited:SetCallback("ChallengeList_OnListSort", CharContainerInvited_OnListSort)
        charContainerInvited:SetDescending(true)
        window.charContainerInvited = charContainerInvited
        -- charContainer / bg
        charContainerInvitedBG:SetPoint("TOPLEFT", charContainerInvited.scrollframe, -5, 5)
        charContainerInvitedBG:SetPoint("BOTTOMRIGHT", charContainerInvited.scrollframe, 5, -5)

        -- register functions
        for k,v in pairs(INV_FRAME_FUNCTIONS) do
            window[k] = v
        end
        INV_FRAME_FUNCTIONS = nil

        GGL:AddCallbackHandler(window)

        window.mergeIDSave = {}
        window.playerList = {}

        GGL.GUI:AddFrameControl("ChallengeInviteFrame", frame, true, true)

        self.inviteFrame = window
    end
    return self.inviteFrame
end

function Challenge:ShowInviteFrame(session)
    Challenge:CreateInviteFrame()
    self.inviteFrame:SetSession(session)
    self.inviteFrame:Show()
    return self.inviteFrame
end